@@ -1,52 +1,76 @@
 use std::borrow::Cow;
 use std::collections::BTreeSet;
+use std::env;
 use std::fmt;
+use std::sync::atomic::{AtomicBool, Ordering};
 
-use clicolors_control;
-use regex::{Matches, Regex};
-use unicode_width::{UnicodeWidthChar, UnicodeWidthStr};
+use crate::term::{wants_emoji, Term};
+use lazy_static::lazy_static;
 
-use term::wants_emoji;
+#[cfg(feature = "ansi-parsing")]
+use crate::ansi::{strip_ansi_codes, AnsiCodeIterator};
 
-lazy_static! {
-    static ref STRIP_ANSI_RE: Regex =
-        Regex::new(r"[\x1b\x9b][\[()#;?]*(?:[0-9]{1,4}(?:;[0-9]{0,4})*)?[0-9A-PRZcf-nqry=><]")
-            .unwrap();
+#[cfg(not(feature = "ansi-parsing"))]
+fn strip_ansi_codes(s: &str) -> &str {
+    s
 }
 
-/// Returns `true` if colors should be enabled.
+fn default_colors_enabled(out: &Term) -> bool {
+    (out.features().colors_supported()
+        && &env::var("CLICOLOR").unwrap_or_else(|_| "1".into()) != "0")
+        || &env::var("CLICOLOR_FORCE").unwrap_or_else(|_| "0".into()) != "0"
+}
+
+lazy_static! {
+    static ref STDOUT_COLORS: AtomicBool = AtomicBool::new(default_colors_enabled(&Term::stdout()));
+    static ref STDERR_COLORS: AtomicBool = AtomicBool::new(default_colors_enabled(&Term::stderr()));
+}
+
+/// Returns `true` if colors should be enabled for stdout.
 ///
 /// This honors the [clicolors spec](http://bixense.com/clicolors/).
 ///
 /// * `CLICOLOR != 0`: ANSI colors are supported and should be used when the program isn't piped.
 /// * `CLICOLOR == 0`: Don't output ANSI color escape codes.
 /// * `CLICOLOR_FORCE != 0`: ANSI colors should be enabled no matter what.
-///
-/// This internally uses `clicolors-control`.
 #[inline]
 pub fn colors_enabled() -> bool {
-    clicolors_control::colors_enabled()
+    STDOUT_COLORS.load(Ordering::Relaxed)
 }
 
-/// Forces colorization on or off.
+/// Forces colorization on or off for stdout.
 ///
 /// This overrides the default for the current process and changes the return value of the
 /// `colors_enabled` function.
-///
-/// This internally uses `clicolors-control`.
 #[inline]
 pub fn set_colors_enabled(val: bool) {
-    clicolors_control::set_colors_enabled(val)
+    STDOUT_COLORS.store(val, Ordering::Relaxed)
 }
 
-/// Helper function to strip ansi codes.
-pub fn strip_ansi_codes(s: &str) -> Cow<str> {
-    STRIP_ANSI_RE.replace_all(s, "")
+/// Returns `true` if colors should be enabled for stderr.
+///
+/// This honors the [clicolors spec](http://bixense.com/clicolors/).
+///
+/// * `CLICOLOR != 0`: ANSI colors are supported and should be used when the program isn't piped.
+/// * `CLICOLOR == 0`: Don't output ANSI color escape codes.
+/// * `CLICOLOR_FORCE != 0`: ANSI colors should be enabled no matter what.
+#[inline]
+pub fn colors_enabled_stderr() -> bool {
+    STDERR_COLORS.load(Ordering::Relaxed)
+}
+
+/// Forces colorization on or off for stderr.
+///
+/// This overrides the default for the current process and changes the return value of the
+/// `colors_enabled` function.
+#[inline]
+pub fn set_colors_enabled_stderr(val: bool) {
+    STDERR_COLORS.store(val, Ordering::Relaxed)
 }
 
 /// Measure the width of a string in terminal characters.
 pub fn measure_text_width(s: &str) -> usize {
-    strip_ansi_codes(s).width()
+    str_width(&strip_ansi_codes(s))
 }
 
 /// A terminal color.
@@ -60,6 +84,7 @@ pub enum Color {
     Magenta,
     Cyan,
     White,
+    Color256(u8),
 }
 
 impl Color {
@@ -74,6 +99,15 @@ impl Color {
             Color::Magenta => 5,
             Color::Cyan => 6,
             Color::White => 7,
+            Color::Color256(x) => x as usize,
+        }
+    }
+
+    #[inline]
+    fn is_color256(self) -> bool {
+        match self {
+            Color::Color256(_) => true,
+            _ => false,
         }
     }
 }
@@ -118,8 +152,11 @@ pub enum Alignment {
 pub struct Style {
     fg: Option<Color>,
     bg: Option<Color>,
+    fg_bright: bool,
+    bg_bright: bool,
     attrs: BTreeSet<Attribute>,
     force: Option<bool>,
+    for_stderr: bool,
 }
 
 impl Default for Style {
@@ -134,8 +171,11 @@ impl Style {
         Style {
             fg: None,
             bg: None,
+            fg_bright: false,
+            bg_bright: false,
             attrs: BTreeSet::new(),
             force: None,
+            for_stderr: false,
         }
     }
 
@@ -157,6 +197,7 @@ impl Style {
                 "magenta" => rv.magenta(),
                 "cyan" => rv.cyan(),
                 "white" => rv.white(),
+                "bright" => rv.bright(),
                 "on_black" => rv.on_black(),
                 "on_red" => rv.on_red(),
                 "on_green" => rv.on_green(),
@@ -165,6 +206,7 @@ impl Style {
                 "on_magenta" => rv.on_magenta(),
                 "on_cyan" => rv.on_cyan(),
                 "on_white" => rv.on_white(),
+                "on_bright" => rv.on_bright(),
                 "bold" => rv.bold(),
                 "dim" => rv.dim(),
                 "underlined" => rv.underlined(),
@@ -196,6 +238,22 @@ impl Style {
         self
     }
 
+    /// Specifies that style is applying to something being written on stderr.
+    #[inline]
+    pub fn for_stderr(mut self) -> Style {
+        self.for_stderr = true;
+        self
+    }
+
+    /// Specifies that style is applying to something being written on stdout.
+    ///
+    /// This is the default behaviour.
+    #[inline]
+    pub fn for_stdout(mut self) -> Style {
+        self.for_stderr = false;
+        self
+    }
+
     /// Sets a foreground color.
     #[inline]
     pub fn fg(mut self, color: Color) -> Style {
@@ -249,6 +307,17 @@ impl Style {
     pub fn white(self) -> Style {
         self.fg(Color::White)
     }
+    #[inline]
+    pub fn color256(self, color: u8) -> Style {
+        self.fg(Color::Color256(color))
+    }
+
+    #[inline]
+    pub fn bright(mut self) -> Style {
+        self.fg_bright = true;
+        self
+    }
+
     #[inline]
     pub fn on_black(self) -> Style {
         self.bg(Color::Black)
@@ -281,6 +350,17 @@ impl Style {
     pub fn on_white(self) -> Style {
         self.bg(Color::White)
     }
+    #[inline]
+    pub fn on_color256(self, color: u8) -> Style {
+        self.bg(Color::Color256(color))
+    }
+
+    #[inline]
+    pub fn on_bright(mut self) -> Style {
+        self.bg_bright = true;
+        self
+    }
+
     #[inline]
     pub fn bold(self) -> Style {
         self.attr(Attribute::Bold)
@@ -348,6 +428,22 @@ impl<D> StyledObject<D> {
         self
     }
 
+    /// Specifies that style is applying to something being written on stderr
+    #[inline]
+    pub fn for_stderr(mut self) -> StyledObject<D> {
+        self.style = self.style.for_stderr();
+        self
+    }
+
+    /// Specifies that style is applying to something being written on stdout
+    ///
+    /// This is the default
+    #[inline]
+    pub fn for_stdout(mut self) -> StyledObject<D> {
+        self.style = self.style.for_stdout();
+        self
+    }
+
     /// Sets a foreground color.
     #[inline]
     pub fn fg(mut self, color: Color) -> StyledObject<D> {
@@ -401,6 +497,17 @@ impl<D> StyledObject<D> {
     pub fn white(self) -> StyledObject<D> {
         self.fg(Color::White)
     }
+    #[inline]
+    pub fn color256(self, color: u8) -> StyledObject<D> {
+        self.fg(Color::Color256(color))
+    }
+
+    #[inline]
+    pub fn bright(mut self) -> StyledObject<D> {
+        self.style = self.style.bright();
+        self
+    }
+
     #[inline]
     pub fn on_black(self) -> StyledObject<D> {
         self.bg(Color::Black)
@@ -433,6 +540,17 @@ impl<D> StyledObject<D> {
     pub fn on_white(self) -> StyledObject<D> {
         self.bg(Color::White)
     }
+    #[inline]
+    pub fn on_color256(self, color: u8) -> StyledObject<D> {
+        self.bg(Color::Color256(color))
+    }
+
+    #[inline]
+    pub fn on_bright(mut self) -> StyledObject<D> {
+        self.style = self.style.on_bright();
+        self
+    }
+
     #[inline]
     pub fn bold(self) -> StyledObject<D> {
         self.attr(Attribute::Bold)
@@ -468,13 +586,32 @@ macro_rules! impl_fmt {
         impl<D: fmt::$name> fmt::$name for StyledObject<D> {
             fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
                 let mut reset = false;
-                if self.style.force.unwrap_or_else(colors_enabled) {
+                if self
+                    .style
+                    .force
+                    .unwrap_or_else(|| match self.style.for_stderr {
+                        true => colors_enabled_stderr(),
+                        false => colors_enabled(),
+                    })
+                {
                     if let Some(fg) = self.style.fg {
-                        write!(f, "\x1b[{}m", fg.ansi_num() + 30)?;
+                        if fg.is_color256() {
+                            write!(f, "\x1b[38;5;{}m", fg.ansi_num())?;
+                        } else if self.style.fg_bright {
+                            write!(f, "\x1b[38;5;{}m", fg.ansi_num() + 8)?;
+                        } else {
+                            write!(f, "\x1b[{}m", fg.ansi_num() + 30)?;
+                        }
                         reset = true;
                     }
                     if let Some(bg) = self.style.bg {
-                        write!(f, "\x1b[{}m", bg.ansi_num() + 40)?;
+                        if bg.is_color256() {
+                            write!(f, "\x1b[48;5;{}m", bg.ansi_num())?;
+                        } else if self.style.bg_bright {
+                            write!(f, "\x1b[48;5;{}m", bg.ansi_num() + 8)?;
+                        } else {
+                            write!(f, "\x1b[{}m", bg.ansi_num() + 40)?;
+                        }
                         reset = true;
                     }
                     for attr in &self.style.attrs {
@@ -533,69 +670,29 @@ impl<'a, 'b> fmt::Display for Emoji<'a, 'b> {
     }
 }
 
-/// An iterator over ansi codes in a string.
-///
-/// This type can be used to scan over ansi codes in a string.
-/// It yields tuples in the form `(s, is_ansi)` where `s` is a slice of
-/// the original string and `is_ansi` indicates if the slice contains
-/// ansi codes or string values.
-pub struct AnsiCodeIterator<'a> {
-    s: &'a str,
-    pending_item: Option<(&'a str, bool)>,
-    last_idx: usize,
-    cur_idx: usize,
-    iter: Matches<'static, 'a>,
-}
-
-impl<'a> AnsiCodeIterator<'a> {
-    /// Creates a new ansi code iterator.
-    pub fn new(s: &'a str) -> AnsiCodeIterator<'a> {
-        AnsiCodeIterator {
-            s,
-            pending_item: None,
-            last_idx: 0,
-            cur_idx: 0,
-            iter: STRIP_ANSI_RE.find_iter(s),
-        }
+fn str_width(s: &str) -> usize {
+    #[cfg(feature = "unicode-width")]
+    {
+        use unicode_width::UnicodeWidthStr;
+        s.width()
     }
-
-    /// Returns the string slice up to the current match.
-    pub fn current_slice(&self) -> &str {
-        &self.s[..self.cur_idx]
-    }
-
-    /// Returns the string slice from the current match to the end.
-    pub fn rest_slice(&self) -> &str {
-        &self.s[self.cur_idx..]
+    #[cfg(not(feature = "unicode-width"))]
+    {
+        s.chars().count()
     }
 }
 
-impl<'a> Iterator for AnsiCodeIterator<'a> {
-    type Item = (&'a str, bool);
-
-    fn next(&mut self) -> Option<(&'a str, bool)> {
-        if let Some(pending_item) = self.pending_item.take() {
-            self.cur_idx += pending_item.0.len();
-            Some(pending_item)
-        } else if let Some(m) = self.iter.next() {
-            let s = &self.s[self.last_idx..m.start()];
-            self.last_idx = m.end();
-            if s.is_empty() {
-                self.cur_idx = m.end();
-                Some((m.as_str(), true))
-            } else {
-                self.cur_idx = m.start();
-                self.pending_item = Some((m.as_str(), true));
-                Some((s, false))
-            }
-        } else if self.last_idx < self.s.len() {
-            let rv = &self.s[self.last_idx..];
-            self.cur_idx = self.s.len();
-            self.last_idx = self.s.len();
-            Some((rv, false))
-        } else {
-            None
-        }
+#[cfg(feature = "ansi-parsing")]
+fn char_width(c: char) -> usize {
+    #[cfg(feature = "unicode-width")]
+    {
+        use unicode_width::UnicodeWidthChar;
+        c.width().unwrap_or(0)
+    }
+    #[cfg(not(feature = "unicode-width"))]
+    {
+        let _c = c;
+        1
     }
 }
 
@@ -606,51 +703,70 @@ impl<'a> Iterator for AnsiCodeIterator<'a> {
 /// escapes code will still be honored.  If truncation takes place
 /// the tail string will be appended.
 pub fn truncate_str<'a>(s: &'a str, width: usize, tail: &str) -> Cow<'a, str> {
-    let mut iter = AnsiCodeIterator::new(s);
-    let mut length = 0;
-    let mut rv = None;
+    #[cfg(feature = "ansi-parsing")]
+    {
+        use std::cmp::Ordering;
+        let mut iter = AnsiCodeIterator::new(s);
+        let mut length = 0;
+        let mut rv = None;
 
-    while let Some(item) = iter.next() {
-        match item {
-            (s, false) => {
-                if rv.is_none() {
-                    if s.width() + length > width - tail.width() {
-                        let ts = iter.current_slice();
+        while let Some(item) = iter.next() {
+            match item {
+                (s, false) => {
+                    if rv.is_none() {
+                        if str_width(s) + length > width - str_width(tail) {
+                            let ts = iter.current_slice();
 
-                        let mut s_byte = 0;
-                        let mut s_width = 0;
-                        let rest_width = width - tail.width() - length;
-                        for c in s.chars() {
-                            s_byte += c.len_utf8();
-                            s_width += c.width().unwrap_or(0);
-                            if s_width == rest_width {
-                                break;
-                            } else if s_width > rest_width {
-                                s_byte -= c.len_utf8();
-                                break;
+                            let mut s_byte = 0;
+                            let mut s_width = 0;
+                            let rest_width = width - str_width(tail) - length;
+                            for c in s.chars() {
+                                s_byte += c.len_utf8();
+                                s_width += char_width(c);
+                                match s_width.cmp(&rest_width) {
+                                    Ordering::Equal => break,
+                                    Ordering::Greater => {
+                                        s_byte -= c.len_utf8();
+                                        break;
+                                    }
+                                    Ordering::Less => continue,
+                                }
                             }
-                        }
 
-                        let idx = ts.len() - s.len() + s_byte;
-                        let mut buf = ts[..idx].to_string();
-                        buf.push_str(tail);
-                        rv = Some(buf);
+                            let idx = ts.len() - s.len() + s_byte;
+                            let mut buf = ts[..idx].to_string();
+                            buf.push_str(tail);
+                            rv = Some(buf);
+                        }
+                        length += str_width(s);
                     }
-                    length += s.width();
                 }
-            }
-            (s, true) => {
-                if rv.is_some() {
-                    rv.as_mut().unwrap().push_str(s);
+                (s, true) => {
+                    if rv.is_some() {
+                        rv.as_mut().unwrap().push_str(s);
+                    }
                 }
             }
         }
+
+        if let Some(buf) = rv {
+            Cow::Owned(buf)
+        } else {
+            Cow::Borrowed(s)
+        }
     }
 
-    if let Some(buf) = rv {
-        Cow::Owned(buf)
-    } else {
-        Cow::Borrowed(s)
+    #[cfg(not(feature = "ansi-parsing"))]
+    {
+        if s.len() <= width - tail.len() {
+            Cow::Borrowed(s)
+        } else {
+            Cow::Owned(format!(
+                "{}{}",
+                s.get(..width - tail.len()).unwrap_or_default(),
+                tail
+            ))
+        }
     }
 }
 
@@ -717,10 +833,18 @@ fn test_text_width() {
         .bold()
         .force_styling(true)
         .to_string();
-    assert_eq!(measure_text_width(&s), 3);
+    assert_eq!(
+        measure_text_width(&s),
+        if cfg!(feature = "ansi-parsing") {
+            3
+        } else {
+            21
+        }
+    );
 }
 
 #[test]
+#[cfg(all(feature = "unicode-width", feature = "ansi-parsing"))]
 fn test_truncate_str() {
     let s = format!("foo {}", style("bar").red().force_styling(true));
     assert_eq!(
@@ -749,6 +873,13 @@ fn test_truncate_str() {
     );
 }
 
+#[test]
+fn test_truncate_str_no_ansi() {
+    assert_eq!(&truncate_str("foo bar", 5, ""), "foo b");
+    assert_eq!(&truncate_str("foo bar", 5, "!"), "foo !");
+    assert_eq!(&truncate_str("foo bar baz", 10, "..."), "foo bar...");
+}
+
 #[test]
 fn test_pad_str() {
     assert_eq!(pad_str("foo", 7, Alignment::Center, None), "  foo  ");
@@ -791,25 +922,3 @@ fn test_pad_str_with() {
         "foo..."
     );
 }
-
-#[test]
-fn test_ansi_iter_re() {
-    let s = format!("Hello {}!", style("World").red().force_styling(true));
-    let mut iter = AnsiCodeIterator::new(&s);
-    assert_eq!(iter.next(), Some(("Hello ", false)));
-    assert_eq!(iter.current_slice(), "Hello ");
-    assert_eq!(iter.rest_slice(), "\x1b[31mWorld\x1b[0m!");
-    assert_eq!(iter.next(), Some(("\x1b[31m", true)));
-    assert_eq!(iter.current_slice(), "Hello \x1b[31m");
-    assert_eq!(iter.rest_slice(), "World\x1b[0m!");
-    assert_eq!(iter.next(), Some(("World", false)));
-    assert_eq!(iter.current_slice(), "Hello \x1b[31mWorld");
-    assert_eq!(iter.rest_slice(), "\x1b[0m!");
-    assert_eq!(iter.next(), Some(("\x1b[0m", true)));
-    assert_eq!(iter.current_slice(), "Hello \x1b[31mWorld\x1b[0m");
-    assert_eq!(iter.rest_slice(), "!");
-    assert_eq!(iter.next(), Some(("!", false)));
-    assert_eq!(iter.current_slice(), "Hello \x1b[31mWorld\x1b[0m!");
-    assert_eq!(iter.rest_slice(), "");
-    assert_eq!(iter.next(), None);
-}
